//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

import SafariServices

import RxSwift
import RxCocoa
import Alamofire
import RxAlamofire

import KakaoSDKCommon
import RxKakaoSDKCommon

import KakaoSDKAuth
import RxKakaoSDKAuth

import KakaoSDKUser
import RxKakaoSDKUser

import KakaoSDKTalk
import RxKakaoSDKTalk

import KakaoSDKFriend
import RxKakaoSDKFriend

import KakaoSDKShare
import RxKakaoSDKShare

import KakaoSDKNavi
import KakaoSDKTemplate

import RxDataSources

// MARK: - tableview helper =============================================================================================

class Row {
    var title: String
    var action: (() -> Void)?
    var apiResult: ApiResult
    init(title: String, action: (() -> Void)?, result: ApiResult = .Ready) {
        self.title = title
        self.action = action
        self.apiResult = result
    }
}

class SectionOfRow : SectionModelType {
    typealias Item = Row
    
    var header: String
    var items: [Item]
    init(header:String, items:[Item]) {
        self.header = header
        self.items = items
    }
    
    required  convenience init(original: SectionOfRow, items: [Item]) {
        self.init(header: original.header, items: original.items)
        self.items = items
    }
}

// MARK: tableview helper =============================================================================================
// MARK: -

extension RxSDKSampleViewController {
    private func updateResultIndicator(cell:UITableViewCell?, row:Row) {
        
        switch(row.apiResult) {
        case .Ready: cell?.backgroundColor = UIColor.white
        case .Success: cell?.backgroundColor = UIColor.green
        case .Fail: cell?.backgroundColor = UIColor.red
        }
    }
    
    private func updateResultIndicator() {
        if let resultIndicator = self.resultIndicator {
            let cell = self.tableView.cellForRow(at: resultIndicator.0)
            self.updateResultIndicator(cell: cell, row: resultIndicator.1)
        }
    }
    
    private func success(_ object: Any? = nil, updateIndicator: Bool = true) {
        if let object = object {
            print("\(String(describing:object))")
        }
        
        if (updateIndicator) {
            DispatchQueue.main.async {
                self.resultIndicator?.1.apiResult = .Success
                self.updateResultIndicator()
            }
        }
    }
    
    
    private func errorHandler(error:Error?, updateIndicator: Bool = true, withAlert: Bool = false) {
        guard error != nil else { return }
        if (updateIndicator) {
            DispatchQueue.main.async {
                self.resultIndicator?.1.apiResult = .Fail
                self.updateResultIndicator()
            }
        }
        
        if let sdkError = error as? SdkError {
            print("\n\n[sdk error handling start]------------------------------------------------------------------------------")
            
            switch(sdkError) {
            case .ApiFailed:
                let apiError = sdkError.getApiError()
                switch(apiError.reason) {
                //if need, use reason case
                default:
                    print("api error: \(String(describing: apiError))")
                }
            case .AuthFailed:
                let authError = sdkError.getAuthError()
                switch(authError.reason) {
                //if need, use reason case
                default:
                    print("auth error: \(String(describing: authError))")
                }
            case .ClientFailed:
                let clientError = sdkError.getClientError()
                switch(clientError.reason) {
                //if need, use reason case
                default:
                    print("client error: \(String(describing: clientError))")
                }
            case .AppsFailed:
                let appsError = sdkError.getAppsError()
                switch(appsError.reason) {
                //if need, use reason case
                default:
                    print("apps error: \(String(describing: appsError))")
                }
            @unknown default:
                break
            }
            
            if withAlert == true {
                self.showAlert(title:"SDK Error", message: String(describing: error))
            }
            print("[sdk error handling end]--------------------------------------------------------------------------------\n")
        }
        else {
            print("\n\n[not sdk error start]------------------------------------------------------------------------------")
            if let nsError = error as NSError? {
                print("normal error: \(String(describing: nsError))")
                
                if withAlert == true {
                    self.showAlert(title:"Normal(Not SDK) Error", message: String(describing: nsError))
                }
            }
            
            print("[not sdk error end]------------------------------------------------------------------------------\n")
        }
    }
}


 class RxSDKSampleViewController: SDKSampleBaseViewController {
    let disposeBag = DisposeBag()
    
    lazy  var dataSource = RxTableViewSectionedReloadDataSource<SectionOfRow>(
        configureCell: { (dataSource, tableView, indexPath, item) in
            let cell = tableView.dequeueReusableCell(withIdentifier: "Cell", for: indexPath)
            
            self.updateResultIndicator(cell:cell, row:item)
            
//            cell.selectionStyle = UITableViewCell.SelectionStyle.none
            cell.textLabel?.adjustsFontSizeToFitWidth = true
            cell.textLabel?.text = "\(item.title)"
            cell.textLabel?.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
            cell.textLabel?.textColor = UIColor.black
            return cell
        })
    
    var safariViewController : SFSafariViewController?
    
    override  func loadView() {
        super.loadView()
        
        setupViews()
    }
    
    override  func viewDidLoad() {
        super.viewDidLoad()
        
        setMenu()
        
        bindViews()
        
        subscribe2()
        
        //COMMENT: 릴리즈모드에서도 subscribe 가능.
//        SdkLog.logger.subscribe{print($0)}
//        .disposed(by: self.disposeBag)
    }
    
    override  func setupViews() {
        super.setupViews()
        self.tableView.register(UITableViewCell.self, forCellReuseIdentifier: "Cell")
        
//        self.navigationItem.rightBarButtonItems = [UIBarButtonItem(title: "Log", style: UIBarButtonItem.Style.plain, target: self, action: #selector(RxSDKSampleViewController.log(sender:)))]
    }
    
//    @objc func log(sender:UIBarButtonItem) {
//        print("logbutton clicked")
//    }
    
    private func bindViews() {
        self.dataSource.titleForHeaderInSection = { dataSource, index in
            return dataSource.sectionModels[index].header
        }
        
        Observable.just(self.sections!)
            .bind(to: tableView.rx.items(dataSource: dataSource))
            .disposed(by: disposeBag)
    }
    
//    func subscribe() {
//        self.tableView.rx.modelSelected(Row.self).subscribe(onNext: { row in
//            if let action = row.action {
//                action()
//            }
//        })
//            .disposed(by: disposeBag)
//    }
    
    func subscribe2() {
        Observable
            .zip(self.tableView.rx.itemSelected, self.tableView.rx.modelSelected(Row.self),
                 resultSelector: { indexPath, row in
                    if let action = row.action {
                        action()
                        
                        self.resultIndicator = (indexPath, row)
                        
                        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1 * 1) {
                           self.tableView.deselectRow(at: indexPath, animated: false)
                        }
                    }
            })
            .subscribe()
            .disposed(by: self.disposeBag)
    }
    
     func setTitle(title:String = "Rx SDK Sample", args: String? = nil) {
        self.navigationItem.title = (args != nil) ? "\(title)(\(args!))" : "\(title)"
    }
    
    func extractPickerElements(friends:Friends<Friend>) -> [(URL?, String?, Any)]? {
        return friends.elements?.map({ (friend) -> (URL?, String?, Any) in return (friend.profileThumbnailImage, friend.profileNickname, friend.uuid) })
    }
    
    func setMenu() {
        self.sections = [
            // MARK: - section auth
            SectionOfRow(header: "AUTH", items: [
                Row(title: "hasToken() usage", action: {
                    if (AuthApi.hasToken()) {
                        UserApi.shared.rx.accessTokenInfo()
                            .subscribe(onSuccess:{ (accessTokenInfo) in
                                self.success("토큰 유효성 체크 성공(필요시 토큰갱신됨)")
                                
                                //do something
                                _ = accessTokenInfo
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    else {
                        print("토큰없음: login 필요")
                    }
                }),
                Row(title: "refreshToken()", action: {
                    AuthApi.shared.rx.refreshToken()
                        .subscribe(onSuccess:{ (oauthToken) in
                            self.success("토큰갱신 성공")
                            
                            //do something
                            _ = oauthToken
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
            ]),
            // MARK: - section user login
            SectionOfRow(header: "USER - login", items: [
                Row(title: "loginWithKakaoTalk()", action: {
                    if (UserApi.isKakaoTalkLoginAvailable()) {
                        UserApi.shared.rx.loginWithKakaoTalk()
                            .subscribe(onNext:{ (oauthToken) in
                            self.success()
                                
                            //do something
                            _ = oauthToken
                                
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title: "loginWithKakaoTalk(launchMethod:) - UniversalLink", action: {
                    UserApi.shared.rx.loginWithKakaoTalk(launchMethod: .UniversalLink)
                            .subscribe(onNext:{ (oauthToken) in
                            self.success()
                                
                            //do something
                            _ = oauthToken
                                
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                
                Row(title: "+loginWithKakaoTalk(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.loginWithKakaoTalk) { paramItems in
                        UserApi.shared.rx.loginWithKakaoTalk(launchMethod: paramItems.requestParam(key:"launchMethod"),
                                                             channelPublicIds: paramItems.requestParam(key:"channelPublicIds"),
                                                             serviceTerms: paramItems.requestParam(key:"serviceTerms"),
                                                             nonce: paramItems.requestParam(key:"nonce"))
                        .subscribe(onNext: { oauthToken in
                            self.success()
                        }, onError: { error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    }
                }),
                
                Row(title: "+loginWithKakaoAccount(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.loginWithKakaoAccount) { paramItems in
                        UserApi.shared.rx.loginWithKakaoAccount(prompts:paramItems.requestParam(key:"prompts"),
                                                                loginHint: paramItems.requestParam(key:"loginHint"),
                                                                nonce: paramItems.requestParam(key:"nonce"))
                        .subscribe(onNext: { oauthToken in
                            self.success()
                            
                            //do something
                            _ = oauthToken
                            
                        }, onError: { error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    }
                    
                }),
            ]),
            
            // MARK: - section user OIDC login
            SectionOfRow(header: "USER - OIDC login", items: [
                Row(title: "me() incremental authorization required", action: {
                    UserApi.shared.rx.me()
                        .map({ (user) -> User in
                            var scopes = [String]()
                            
                            if (user.kakaoAccount?.emailNeedsAgreement == true) { scopes.append("account_email") }
                            if (user.kakaoAccount?.birthdayNeedsAgreement == true) { scopes.append("birthday") }
                            if (user.kakaoAccount?.birthyearNeedsAgreement == true) { scopes.append("birthyear") }
                            if (user.kakaoAccount?.ciNeedsAgreement == true) { scopes.append("account_ci") }
                            if (user.kakaoAccount?.phoneNumberNeedsAgreement == true) { scopes.append("phone_number") }
                            if (user.kakaoAccount?.profileNeedsAgreement == true) { scopes.append("profile") }
                            if (user.kakaoAccount?.ageRangeNeedsAgreement == true) { scopes.append("age_range") }
                            
                            if (scopes.count > 0) {
                                
                                // OpenID 활성화 후
                                // - scope 파라메터에 openid 항목을 포함하여 요청할 경우 OIDC로 동작
                                // - scope 파라메터에 openid 항목을 미포함 시 OAuth2.0 으로 동작
                                                                 
                                // OIDC 요청이므로 "openid" 항목을 추가한다
                                scopes.append("openid")
                                
                                throw SdkError(scopes:scopes)
                            }
                            else {
                                return user
                            }
                        })
                        .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired(nonce:"test_nonce"))
                        .subscribe(onSuccess:{ ( user ) in
                            self.success()
                            
                            //do something
                            _ = user
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
            ]),            
            // MARK: - section user other api
            SectionOfRow(header: "USER - other api", items: [
                Row(title: "signup()", action: {
                    UserApi.shared.rx.signup()
                        .subscribe (onSuccess:{ (userId) in
                            self.success()
                            
                            let _ = userId
                            //do something
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"logout()", action: {
                    UserApi.shared.rx.logout()
                        .subscribe(onCompleted:{
                            self.success("onCompleted:")
                            //do something
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"unlink()", action: {
                    UserApi.shared.rx.unlink()
                        .subscribe(onCompleted:{
                            self.success("onCompleted:")
                            //do something
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "me()", action: {
                    UserApi.shared.rx.me()
                        .subscribe (onSuccess:{ user in
                            self.success()
                            
                            //do something
                            _ = user
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    
//                    UserApi.shared.rx.me(propertyKeys:["id"])
//                        .subscribe (onSuccess:{ user in
//                            self.success()
//                            
//                            //do something
//                            _ = user
//                            
//                        }, onError: {error in
//                            self.errorHandler(error: error)
//                        })
//                        .disposed(by: self.disposeBag)
                }),
                Row(title: "me() incremental authorization required", action: {
                    UserApi.shared.rx.me()
                        .map({ (user) -> User in
                            var scopes = [String]()
                            
                            if (user.kakaoAccount?.emailNeedsAgreement == true) { scopes.append("account_email") }
                            if (user.kakaoAccount?.birthdayNeedsAgreement == true) { scopes.append("birthday") }
                            if (user.kakaoAccount?.birthyearNeedsAgreement == true) { scopes.append("birthyear") }
                            if (user.kakaoAccount?.ciNeedsAgreement == true) { scopes.append("account_ci") }
                            if (user.kakaoAccount?.phoneNumberNeedsAgreement == true) { scopes.append("phone_number") }
                            if (user.kakaoAccount?.profileNeedsAgreement == true) { scopes.append("profile") }
                            if (user.kakaoAccount?.ageRangeNeedsAgreement == true) { scopes.append("age_range") }
                            
                            if (scopes.count > 0) {
                                throw SdkError(scopes:scopes)
                            }
                            else {
                                return user
                            }
                        })
                        .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess:{ ( user ) in
                            self.success()
                            
                            //do something
                            _ = user
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"updateProfile()", action: {
                    UserApi.shared.rx.updateProfile(properties:["custom_key":"update_profile_test"])
                        .subscribe(onCompleted:{
                            self.success()
                            
                            //do something
                                                        
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"accessTokenInfo()", action: {
                    UserApi.shared.rx.accessTokenInfo()
                        .subscribe(onSuccess:{ (accessTokenInfo) in
                            self.success()
                            
                            //do something
                            _ = accessTokenInfo
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"shippingAddresses(fromUpdatedAt:pageSize:)", action: {
                    UserApi.shared.rx.shippingAddresses()
                        .subscribe(onSuccess:{ (shippingAddress) in
                            self.success()
                            
                            //do something
                            _ = shippingAddress
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "+serviceTerms(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.serviceTerms) { paramItems in
                        UserApi.shared.rx.serviceTerms(result:paramItems.requestParam(key:"result"), tags: paramItems.requestParam(key: "tags"))
                            .subscribe(onSuccess: { serviceTerms in
                                self.success(serviceTerms)
                            }, onFailure: { error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"+revokeServiceTerms(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.revokeServiceTerms) { paramItems in
                        UserApi.shared.rx.revokeServiceTerms(tags: paramItems.requestParam(key: "serviceTerms") ?? [])
                            .subscribe(onSuccess: { (revokeTerms) in
                                self.success(revokeTerms)
                            }, onFailure: { error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"+scopes(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.scopes) { paramItems in
                        UserApi.shared.rx.scopes(scopes: nil)
                            .subscribe(onSuccess:{ (scopeInfo) in
                                self.success(scopeInfo)
                                
                                //do something
                                _ = scopeInfo
                                                            
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"+revokeScopes(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.revokeScopes) { paramItems in
                        UserApi.shared.rx.revokeScopes(scopes: paramItems.requestParam(key:"scopes") ?? [])
                            .subscribe(onSuccess:{ (scopeInfo) in
                                self.success(scopeInfo)
                                
                                //do something
                                _ = scopeInfo
                                                            
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title: "selectShippingAddress(...)", action: {
                    UserApi.shared.rx.selectShippingAddress()
                        .subscribe { addressId in
                            self.success(addressId)
                        } onFailure: { error in
                            self.errorHandler(error: error)
                        }
                        .disposed(by: self.disposeBag)

                })
            ]),
            
            // MARK: - section talk
            SectionOfRow(header: "TALK", items: [
              Row(title: "profile()", action: {
                  TalkApi.shared.rx.profile()
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe (onSuccess:{ (profile) in
                          self.success()
                          
                          //do something
                          _ = profile
                          
                      }, onFailure: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
              }),
              Row(title: "friends()", action: {
                  TalkApi.shared.rx.friends()
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe (onSuccess:{ (friends) in
                          self.success()
                          
                          //do something
                          _ = friends
                          
                      }, onFailure: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
              }),
              Row(title: "+friends(...)", action: {
                  self.showParamDialog(apiKey: ApiKey.friends) { paramItems in
                      TalkApi.shared.rx.friends(offset: paramItems.requestParam(key:"offset"),
                                                limit: paramItems.requestParam(key:"limit"),
                                                order: paramItems.requestParam(key:"order"),
                                                friendOrder: paramItems.requestParam(key:"friendOrder"))
                      .subscribe { friends in
                          self.success(friends)
                      } onFailure: { error in
                          self.errorHandler(error: error)
                      }
                      .disposed(by: self.disposeBag)

                  }
              }),
              Row(title: "friends(offset:limit:order:)", action: {
                  TalkApi.shared.rx.friends(offset: 0, limit: 1, order: .Desc).retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                  .subscribe (onSuccess:{ (friends) in
                      self.success()
                      
                      //do something
                      _ = friends
                      
                  }, onFailure: {error in
                      self.errorHandler(error: error)
                  })
                  .disposed(by: self.disposeBag)
              }),
              Row(title: "+channels(...)", action: {
                  self.showParamDialog(apiKey: ApiKey.channels) { paramItems in
                      TalkApi.shared.rx.channels(publicIds: paramItems.requestParam(key:"publicIds"))
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe (onSuccess:{ (channels) in
                              self.success()
                              
                              //do something
                              _ = channels
                              
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              Row(title: "+followChannel(...)", action: {
                  self.showParamDialog(apiKey: ApiKey.followChannel, completionDelay: 0.4) { paramItems in
                      let channelPublicId:String? = paramItems.requestParam(key:"channelPublicId")
                      if let channelPublicId = channelPublicId {
                          TalkApi.shared.rx.followChannel(channelPublicId: channelPublicId)
                              .subscribe(onSuccess: { (result) in
                                  self.success(result)
                              }, onFailure: { error in
                                  self.errorHandler(error: error)
                              })
                              .disposed(by: self.disposeBag)
                      }
                  }
              }),
              
              //MARK: 자기자신에게 send memo
              Row(title: "sendDefaultMemo(templatable:)", action: {
                
                  let feedTemplateJsonStringData =
                    """
                    {
                        "object_type": "feed",
                        "content": {
                            "title": "딸기 치즈 케익",
                            "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                            "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                            "link": {
                                "mobile_web_url": "https://developers.kakao.com",
                                "web_url": "https://developers.kakao.com"
                            }
                        },
                        "item_content" : {
                            "profile_text" :"Kakao",
                            "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                            "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                            "title_image_text" :"Cheese cake",
                            "title_image_category" : "cake",
                            "items" : [
                                          {
                                              "item" :"cake1",
                                              "item_op" : "1000원",
                                          },
                                          {
                                              "item" :"cake2",
                                              "item_op" : "2000원",
                                          },
                                          {
                                              "item" :"cake3",
                                              "item_op" : "3000원",
                                          },
                                          {
                                              "item" :"cake4",
                                              "item_op" : "4000원",
                                          },
                                          {
                                              "item" :"cake5",
                                              "item_op" : "5000원",
                                          }
                                      ],
                                      "sum" :"total",
                                      "sum_op" : "15000원"
                        },
                        "social": {
                            "comment_count": 45,
                            "like_count": 286,
                            "shared_count": 845
                        },
                        "buttons": [
                            {
                                "title": "웹으로 보기",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            },
                            {
                                "title": "앱으로 보기",
                                "link": {
                                    "android_execution_params": "key1=value1&key2=value2",
                                    "ios_execution_params": "key1=value1&key2=value2"
                                }
                            }
                        ]
                    }
                    """.data(using: .utf8)!

                  //TODO: 디코드 메소드 래핑
                  if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                      print("Template URL decoding test: \(templatable.content.imageUrl.absoluteString)")
                      TalkApi.shared.rx.sendDefaultMemo(templatable:templatable)
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe (onCompleted:{
                              self.success()
                              
                              //do something
                              
                          }, onError: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              Row(title: "sendCustomMemo(templateId:)", action: {
                  TalkApi.shared.rx.sendCustomMemo(templateId: AppDelegate.templateId().message)
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe (onCompleted:{
                          self.success()
                          
                          //do something
                          
                      }, onError: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
              }),
              Row(title: "sendCustomMemo(templateId:templateArgs:)", action: {
                  TalkApi.shared.rx.sendCustomMemo(templateId: AppDelegate.templateId().message, templateArgs: ["msg":"메세지 파라미터 테스트입니다."])
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe (onCompleted:{
                          self.success()
                          
                          //do something
                          
                      }, onError: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
              }),
              Row(title: "sendScrapMemo(requestUrl:)", action: {
                  TalkApi.shared.rx.sendScrapMemo(requestUrl: "https://developers.kakao.com")
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe (onCompleted:{
                          self.success()
                          
                          //do something
                          
                      }, onError: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
              }),
              
              //MARK: 톡 친구에게 send message - 디폴트메세지
              Row(title: "sendDefaultMessage(templatable:receiverUuids:) - feed", action: {
                  
                  let feedTemplateJsonStringData =
                    """
                    {
                        "object_type": "feed",
                        "content": {
                            "title": "딸기 치즈 케익",
                            "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                            "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                            "link": {
                                "mobile_web_url": "https://developers.kakao.com",
                                "web_url": "https://developers.kakao.com"
                            }
                        },
                        "item_content" : {
                            "profile_text" :"Kakao",
                            "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                            "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                            "title_image_text" :"Cheese cake",
                            "title_image_category" : "cake",
                            "items" : [
                                          {
                                              "item" :"cake1",
                                              "item_op" : "1000원",
                                          },
                                          {
                                              "item" :"cake2",
                                              "item_op" : "2000원",
                                          },
                                          {
                                              "item" :"cake3",
                                              "item_op" : "3000원",
                                          },
                                          {
                                              "item" :"cake4",
                                              "item_op" : "4000원",
                                          },
                                          {
                                              "item" :"cake5",
                                              "item_op" : "5000원",
                                          }
                                      ],
                                      "sum" :"total",
                                      "sum_op" : "15000원"
                        },
                        "social": {
                            "comment_count": 45,
                            "like_count": 286,
                            "shared_count": 845
                        },
                        "buttons": [
                            {
                                "title": "웹으로 보기",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            },
                            {
                                "title": "앱으로 보기",
                                "link": {
                                    "android_execution_params": "key1=value1&key2=value2",
                                    "ios_execution_params": "key1=value1&key2=value2"
                                }
                            }
                        ]
                    }
                    """.data(using: .utf8)!

                  if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult

                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              
              Row(title:"sendDefaultMessage(templatable:receiverUuids:) - list", action: {
                  
                  let listTemplateJsonStringData =
                    """
                    {
                        "object_type": "list",
                        "header_title": "WEEKLY MAGAZINE",
                        "header_link": {
                            "mobile_web_url": "https://developers.kakao.com",
                            "web_url": "https://developers.kakao.com"
                        },
                        "contents": [
                            {
                                "title": "취미의 특징, 탁구",
                                "description": "스포츠",
                                "image_url": "http://mud-kage.kakao.co.kr/dn/bDPMIb/btqgeoTRQvd/49BuF1gNo6UXkdbKecx600/kakaolink40_original.png",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            },
                            {
                                "title": "크림으로 이해하는 커피이야기",
                                "description": "음식",
                                "image_url": "http://mud-kage.kakao.co.kr/dn/QPeNt/btqgeSfSsCR/0QJIRuWTtkg4cYc57n8H80/kakaolink40_original.png",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            },
                            {
                                "title": "감성이 가득한 분위기",
                                "description": "사진",
                                "image_url": "http://mud-kage.kakao.co.kr/dn/c7MBX4/btqgeRgWhBy/ZMLnndJFAqyUAnqu4sQHS0/kakaolink40_original.png",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            }
                        ],
                        "buttons": [
                            {
                                "title": "웹으로 보기",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            },
                            {
                                "title": "앱으로 보기",
                                "link": {
                                    "android_execution_params": "key1=value1&key2=value2",
                                    "ios_execution_params": "key1=value1&key2=value2"
                                }
                            }
                        ]
                    }
                    """.data(using: .utf8)!

                  if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult
                                                                                      
                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              Row(title:"sendDefaultMessage(templatable:receiverUuids:) - location", action: {
                  
                  let locationTemplateJsonStringData =
                    """
                    {
                        "object_type": "location",
                        "address_title": "카카오 판교오피스 카페톡",
                        "address": "경기 성남시 분당구 판교역로 235 에이치스퀘어 N동 8층",
                        "content": {
                            "description": "이번 주는 체리블라썸라떼 1+1",
                            "image_url": "http://mud-kage.kakao.co.kr/dn/bSbH9w/btqgegaEDfW/vD9KKV0hEintg6bZT4v4WK/kakaolink40_original.png",
                            "link": {
                                "mobile_web_url": "https://developers.kakao.com",
                                "web_url": "https://developers.com"
                            },
                            "title": "신메뉴 출시❤️ 체리블라썸라떼"
                        },
                        "social": {
                            "comment_count": 45,
                            "like_count": 286,
                            "shared_count": 845
                        }
                    }
                    """.data(using: .utf8)!
                  
                  if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult
                                                                                      
                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
                  
              }),
              Row(title:"sendDefaultMessage(templatable:receiverUuids:) - commerce", action: {
                  
                  let commerceTemplateJsonStringData =
                    """
                    {
                        "object_type": "commerce",
                        "commerce": {
                            "product_name": "Ivory long dress",
                            "regular_price": 208800,
                            "discount_price": 146160,
                            "discount_rate": 30,
                            "currency_unit": "₩",
                            "currency_unit_position": 1
                        },
                        "content": {
                            "title": "Ivory long dress (4 Color)",
                            "image_url": "http://mud-kage.kakao.co.kr/dn/RY8ZN/btqgOGzITp3/uCM1x2xu7GNfr7NS9QvEs0/kakaolink40_original.png",
                            "link": {
                                "mobile_web_url": "https://developers.kakao.com",
                                "web_url": "https://developers.kakao.com"
                            },
                        },
                        "buttons": [
                            {
                                "title": "구매하기",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            },
                            {
                                "title": "공유하기",
                                "link": {
                                    "android_execution_params": "key1=value1&key2=value2",
                                    "ios_execution_params": "key1=value1&key2=value2"
                                }
                            }
                        ]
                    }
                    """.data(using: .utf8)!
                  
                  if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult
                                                                                      
                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              Row(title:"sendDefaultMessage(templatable:receiverUuids:) - text", action: {
                  let text = "카카오링크는 카카오 플랫폼 서비스의 대표 기능으로써 사용자의 모바일 기기에 설치된 카카오 플랫폼과 연동하여 다양한 기능을 실행할 수 있습니다.\\n\\n현재 이용할 수 있는 카카오링크는 다음과 같습니다.\\n카카오톡링크\\n카카오톡을 실행하여 사용자가 선택한 채팅방으로 메시지를 전송합니다.\\n카카오스토리링크\\n카카오스토리 글쓰기 화면으로 연결합니다."
                  let textTemplateJsonStringData =
                  """
                  {
                      "object_type": "text",
                      "text": "\(text)",
                      "link": {
                          "web_url": "https://developers.kakao.com",
                          "mobile_web_url": "https://developers.kakao.com"
                      },
                      "button_title": "바로 확인"
                  }
                  """.data(using: .utf8)!
                  
                  if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult
                                                                                  
                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              Row(title:"sendDefaultMessage(templatable:receiverUuids:) - calendar", action: {

                  let calendarTemplateJsonStringData =
                    """
                    {
                        "object_type": "calendar",
                        "id_type": "\(CalendarTemplate.IdType.Calendar.rawValue)",
                        "id": "\(AppDelegate.calendarId().calendar)",
                        "content": {
                            "link": {
                                "web_url": "https://devtalk.kakao.com",
                                "mobile_web_url": "https://developers.kakao.com"
                            },
                            "title": "2022 카타르 월드컵",
                            "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                        },
                        "buttons": [
                            {
                                "title": "월드컵 일정 보기",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            }
                        ]
                    }
                    """.data(using: .utf8)!

                  if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult
                                                                                  
                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              
              Row(title:"sendDefaultMessage(templatable:receiverUuids:) - calendar(event)", action: {

                  let calendarTemplateJsonStringData =
                    """
                    {
                        "object_type": "calendar",
                        "id_type": "\(CalendarTemplate.IdType.Event.rawValue)",
                        "id": "\(AppDelegate.calendarId().event)",
                        "content": {
                            "link": {
                                "web_url": "https://devtalk.kakao.com",
                                "mobile_web_url": "https://developers.kakao.com"
                            },
                            "title": "1월 신작 평론 모임",
                            "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                        },
                        "buttons": [
                            {
                                "title": "모임 주제 보기",
                                "link": {
                                    "mobile_web_url": "https://developers.kakao.com",
                                    "web_url": "https://developers.kakao.com"
                                }
                            }
                        ]
                    }
                    """.data(using: .utf8)!
                  
                  if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                      TalkApi.shared.rx.friends()
                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                          .subscribe(onSuccess: { (friends) in
                              self.success(friends, updateIndicator: false)
                              
                              if let elements = self.extractPickerElements(friends: friends) {
                                  let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                           elements: elements,
                                                                           elementsPickerCompletionHandler: { (selectedIds) in
                                                                              let friendUuids = selectedIds as! [String]
                                                                              TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                          .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                                  .subscribe(onSuccess:{ (messageSendResult) in
                                                                                      self.success()
                                                                                      
                                                                                      //do something
                                                                                      _ = messageSendResult
                                                                                  
                                                                                  }, onFailure: {error in
                                                                                      self.errorHandler(error: error)
                                                                                  })
                                                                                  .disposed(by: self.disposeBag)
                                                                              
                                  })
                                  self.navigationController?.pushViewController(vc, animated: true)
                              }
                          }, onFailure: {error in
                              self.errorHandler(error: error)
                          })
                          .disposed(by: self.disposeBag)
                  }
              }),
              
              //MARK: 톡 친구에게 send message - 커스텀메세지
              Row(title: "sendCustomMessage(templateId:templateArgs:receiverUuids:)", action: {
                  TalkApi.shared.rx.friends()
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe(onSuccess: { (friends) in
                          self.success(friends, updateIndicator: false)
                          
                          if let elements = self.extractPickerElements(friends: friends) {
                              let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                       elements: elements,
                                                                       elementsPickerCompletionHandler: { (selectedIds) in
                                                                          let friendUuids = selectedIds as! [String]
                                                                          TalkApi.shared.rx.sendCustomMessage(templateId:AppDelegate.templateId().message, templateArgs:["title":"제목입니다.", "description":"설명입니다."], receiverUuids:friendUuids)
                                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                              .subscribe(onSuccess:{ (messageSendResult) in
                                                                                  self.success()
                                                                                  
                                                                                  //do something
                                                                                  _ = messageSendResult

                                                                              }, onFailure: {error in
                                                                                  self.errorHandler(error: error)
                                                                              })
                                                                              .disposed(by: self.disposeBag)
                                                                          
                              })
                              self.navigationController?.pushViewController(vc, animated: true)
                          }
                      }, onFailure: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
              }),
              
              //MARK: 톡 친구에게 send message - 스크랩메세지
              Row(title: "sendScrapMessage(requestUrl:receiverUuids:)", action: {
                  TalkApi.shared.rx.friends()
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe(onSuccess: { (friends) in
                          self.success(friends, updateIndicator: false)
                          
                          if let elements = self.extractPickerElements(friends: friends) {
                              let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                       elements: elements,
                                                                       elementsPickerCompletionHandler: { (selectedIds) in
                                                                          let friendUuids = selectedIds as! [String]
                                                                          TalkApi.shared.rx.sendScrapMessage(requestUrl:"https://developers.kakao.com", receiverUuids:friendUuids)
                                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                              .subscribe(onSuccess:{ (messageSendResult) in
                                                                                  self.success()
                                                                                  
                                                                                  //do something
                                                                                  _ = messageSendResult

                                                                              }, onFailure: {error in
                                                                                  self.errorHandler(error: error)
                                                                              })
                                                                              .disposed(by: self.disposeBag)
                                                                          
                              })
                              self.navigationController?.pushViewController(vc, animated: true)
                          }
                      }, onFailure: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
                  
              }),
              Row(title: "sendScrapMessage(requestUrl:templateId:receiverUuids:)", action: {
                  TalkApi.shared.rx.friends()
                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                      .subscribe(onSuccess: { (friends) in
                          self.success(friends, updateIndicator: false)
                          
                          if let elements = self.extractPickerElements(friends: friends) {
                              let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                       elements: elements,
                                                                       elementsPickerCompletionHandler: { (selectedIds) in
                                                                          let friendUuids = selectedIds as! [String]
                                                                          TalkApi.shared.rx.sendScrapMessage(requestUrl:"https://developers.kakao.com", templateId:AppDelegate.templateId().message, receiverUuids:friendUuids)
                                      .retry(when: AuthApiCommon.shared.rx.incrementalAuthorizationRequired())
                                                                              .subscribe(onSuccess:{ (messageSendResult) in
                                                                                 self.success()
                                                                                 
                                                                                 //do something
                                                                                 _ = messageSendResult
                                                                                  
                                                                              }, onFailure: {error in
                                                                                  self.errorHandler(error: error)
                                                                              })
                                                                              .disposed(by: self.disposeBag)
                                                                          
                              })
                              self.navigationController?.pushViewController(vc, animated: true)
                          }
                      }, onFailure: {error in
                          self.errorHandler(error: error)
                      })
                      .disposed(by: self.disposeBag)
                  
              }),
              
              //MARK: 카카오톡 채널
              Row(title: "+addChannel(...) - web", action: {
                  self.showParamDialog(apiKey: ApiKey.addChannel) { paramItems in
                      let channelId: String? = paramItems.requestParam(key:"channelPublicId")
                      if let channelId = channelId {
                          self.safariViewController = SFSafariViewController(url: TalkApi.shared.makeUrlForAddChannel(channelPublicId:channelId)!)
                          self.safariViewController?.modalTransitionStyle = .crossDissolve
                          self.safariViewController?.modalPresentationStyle = .overCurrentContext
                          self.present(self.safariViewController!, animated: true) {
                              print("add plusfriend present success")
                          }
                      }
                  }
              }),
              Row(title: "+chatChannel(...) - web", action: {
                  self.showParamDialog(apiKey: ApiKey.chatChannel) { paramItems in
                      let channelId: String? = paramItems.requestParam(key: "channelPublicId")
                      if let channelId = channelId {
                          self.safariViewController = SFSafariViewController(url: TalkApi.shared.makeUrlForChatChannel(channelPublicId:channelId)!)
                          self.safariViewController?.modalTransitionStyle = .crossDissolve
                          self.safariViewController?.modalPresentationStyle = .overCurrentContext
                          self.present(self.safariViewController!, animated: true) {
                              print("chat present success")
                          }
                      }
                  }
              }),
              Row(title: "+addChannel(...) - scheme", action: {
                  self.showParamDialog(apiKey: ApiKey.addChannel) { paramItems in
                      let channelId: String? = paramItems.requestParam(key: "channelPublicId")
                      if let channelId = channelId {
                          TalkApi.shared.rx.addChannel(channelPublicId: channelId)
                              .subscribe(onCompleted:{
                                  self.success("onCompleted:")
                                  //do something
                              }, onError: {error in
                                  self.errorHandler(error: error)
                              })
                              .disposed(by: self.disposeBag)
                      }
                  }
              }),
              Row(title: "+chatChannel(...) - scheme", action: {
                  self.showParamDialog(apiKey: ApiKey.chatChannel) { paramItems in
                      let channelId: String? = paramItems.requestParam(key: "channelPublicId")
                      if let channelId = channelId {
                          TalkApi.shared.rx.chatChannel(channelPublicId: channelId)
                              .subscribe(onCompleted:{
                                  self.success("onCompleted:")
                                  //do something
                              }, onError: {error in
                                  self.errorHandler(error: error)
                              })
                              .disposed(by: self.disposeBag)
                      }
                  }
              })              
            ]),
            // MARK: - section FRIEND
            SectionOfRow(header: "FRIEND", items: [
                
                Row(title: "selectFriends(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams(title: "샘플앱 친구피커",                     //default "카카오톡 친구 선택"
                                                                    viewAppearance: .auto,                   //default .auto
                                                                    orientation: .auto,                     //default .auto
                                                                    enableSearch: true,                     //default true
                                                                    enableIndex: true,                      //default true
                                                                    showMyProfile: true,                    //default true
                                                                    showFavorite: true,                     //default true
                                                                    showPickedFriend: nil,                  //default true
                                                                    maxPickableCount: nil,                  //default 30
                                                                    minPickableCount: nil)                  //default 1
                    
                    PickerApi.shared.rx.selectFriends(params:openPickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriendsPopup(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams()
                    
                    PickerApi.shared.rx.selectFriendsPopup(params:openPickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriend(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams(title: "샘플앱 친구피커",                     //default "카카오톡 친구 선택"
                                                                    viewAppearance: .auto,                  //default .auto
                                                                    orientation: .auto,                     //default .auto
                                                                    enableSearch: true,                     //default true
                                                                    enableIndex: true,                      //default true
                                                                    showMyProfile: true,                    //default true
                                                                    showFavorite: true,                     //default true
                                                                    showPickedFriend: nil,                  //ignore
                                                                    maxPickableCount: nil,                  //ignore
                                                                    minPickableCount: nil)                  //ignore
                    
                    PickerApi.shared.rx.selectFriend(params:openPickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriendPopup(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams()
                    
                    PickerApi.shared.rx.selectFriendPopup(params:openPickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
            ]),
            // MARK: - section kakaoshare
            SectionOfRow(header: "KAKAOSHARE", items: [
                Row(title: "shareDefault(templatable:) - webSharer", action: {
                    
                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                              "profile_text" :"Kakao",
                              "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "title_image_text" :"Cheese cake",
                              "title_image_category" : "cake",
                              "items" : [
                                            {
                                                "item" :"cake1",
                                                "item_op" : "1000원",
                                            },
                                            {
                                                "item" :"cake2",
                                                "item_op" : "2000원",
                                            },
                                            {
                                                "item" :"cake3",
                                                "item_op" : "3000원",
                                            },
                                            {
                                                "item" :"cake4",
                                                "item_op" : "4000원",
                                            },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        if let url = ShareApi.shared.makeDefaultUrl(templatable: templatable) {
                            print("=============================sharer default url: \(url)")
                            self.safariViewController = SFSafariViewController(url: url)
                            self.safariViewController?.modalTransitionStyle = .crossDissolve
                            self.safariViewController?.modalPresentationStyle = .overCurrentContext
                            self.present(self.safariViewController!, animated: true) {
                                print("웹 present success")
                            }
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:) - webSharer", action: {
                    if let url = ShareApi.shared.makeScrapUrl(requestUrl: "https://developers.kakao.com") {
                        print("=============================sharer scrap url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:templateId:) - webSharer", action: {
                    if let url = ShareApi.shared.makeScrapUrl(requestUrl: "https://developers.kakao.com", templateId: AppDelegate.templateId().link ) {
                        print("=============================sharer scrap url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareCustom(templateId:,templateArgs:) - webSharer", action: {
                    if let url = ShareApi.shared.makeCustomUrl(templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."]) {
                        print("=============================sharer custom url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:)", action: {
                    ShareApi.shared.rx.shareScrap(requestUrl: "https://developers.kakao.com")
                        .subscribe(onSuccess: { (sharingResult) in
                            self.success()
                            
                            //do something
                            UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "shareScrap(requestUrl:templateId:)", action: {
                    ShareApi.shared.rx.shareScrap(requestUrl: "https://developers.kakao.com", templateId: AppDelegate.templateId().link)
                        .subscribe(onSuccess: { (sharingResult) in
                            self.success()
                            
                            //do something
                            UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "shareCustom(templateId:templateArgs:)", action: {
                    ShareApi.shared.rx.shareCustom(templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."])
                        .subscribe(onSuccess: { (sharingResult) in
                            self.success()
                            
                            //do something
                            UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            
                        }, onFailure: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"shareDefault(templatable:serverCallbackArgs:)", action: {
                    
                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                    "profile_text" :"Kakao",
                                    "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                    "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                    "title_image_text" :"Cheese cake",
                                    "title_image_category" : "cake",
                                    "items" : [
                                                  {
                                                      "item" :"cake1",
                                                      "item_op" : "1000원",
                                                  },
                                                  {
                                                      "item" :"cake2",
                                                      "item_op" : "2000원",
                                                  },
                                                  {
                                                      "item" :"cake3",
                                                      "item_op" : "3000원",
                                                  },
                                                  {
                                                      "item" :"cake4",
                                                      "item_op" : "4000원",
                                                  },
                                                  {
                                                      "item" :"cake5",
                                                      "item_op" : "5000원",
                                                  }
                                              ],
                                              "sum" :"total",
                                              "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!
                    
                    let serverCallbackArgs = ["user_id": "efgh", "product_id": "5678"]

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable, serverCallbackArgs:serverCallbackArgs)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"shareDefault(templateObject:) using json", action: {
                    
                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                    "profile_text" :"Kakao",
                                    "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                    "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                    "title_image_text" :"Cheese cake",
                                    "title_image_category" : "cake",
                                    "items" : [
                                                  {
                                                      "item" :"cake1",
                                                      "item_op" : "1000원",
                                                  },
                                                  {
                                                      "item" :"cake2",
                                                      "item_op" : "2000원",
                                                  },
                                                  {
                                                      "item" :"cake3",
                                                      "item_op" : "3000원",
                                                  },
                                                  {
                                                      "item" :"cake4",
                                                      "item_op" : "4000원",
                                                  },
                                                  {
                                                      "item" :"cake5",
                                                      "item_op" : "5000원",
                                                  }
                                              ],
                                              "sum" :"total",
                                              "sum_op" : "15000"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    //jsonObject
                    if let feedTemplateJsonObject = SdkUtils.toJsonObject(feedTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templateObject:feedTemplateJsonObject)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"shareDefault(templateObject:) using constructor", action: {

                    let link = Link(webUrl: URL(string:"https://developers.kakao.com"),
                                    mobileWebUrl: URL(string:"https://developers.kakao.com"))
                    let appLink = Link(androidExecutionParams: ["key1": "value1", "key2": "value2"],
                                       iosExecutionParams: ["key1": "value1", "key2": "value2"])
                    let button1 = Button(title: "웹으로 보기", link: link)
                    let button2 = Button(title: "앱으로 보기", link: appLink)
                    
                    let itemContent = ItemContent(profileText: "Kakao",
                                    profileImageUrl: URL(string:"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png"),
                                    titleImageText: "Cheese cake",
                                    titleImageUrl: URL(string:"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png"),
                                    titleImageCategory: "cake",
                                    items: [ItemInfo(item:"cake1", itemOp:"1000원"), ItemInfo(item:"cake2", itemOp:"2000원"),
                                               ItemInfo(item:"cake3", itemOp:"3000원"), ItemInfo(item:"cake4", itemOp:"4000원"),
                                               ItemInfo(item:"cake5", itemOp:"5000원")],
                                    sum: "total",
                                    sumOp: "15000원")
                    
                    let social = Social(likeCount: 286,
                                        commentCount: 45,
                                        sharedCount: 845)
                    let content = Content(title: "딸기 치즈 케익",
                                          imageUrl: URL(string:"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png")!,
                                          description: "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                                          link: link)
                    let feedTemplate = FeedTemplate(content: content, itemContent: itemContent, social: social, buttons: [button1, button2])
                    
                    //encode 는 이렇게 한다.
                    if let feedTemplateJsonData = (try? SdkJSONEncoder.custom.encode(feedTemplate)) {
                    
                    //jsonObject
                        if let templateJsonObject = SdkUtils.toJsonObject(feedTemplateJsonData) {
                            ShareApi.shared.rx.shareDefault(templateObject:templateJsonObject)
                                .subscribe(onSuccess: { (sharingResult) in
                                    self.success()
                                    
                                    //do something
                                    UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                    
                                }, onFailure: {error in
                                    self.errorHandler(error: error)
                                })
                                .disposed(by: self.disposeBag)
                        }
                    }
                }),
                Row(title:"shareDefault(templatable:) - list", action: {
                    
                    let listTemplateJsonStringData =
                      """
                      {
                          "object_type": "list",
                          "header_title": "WEEKLY MAGAZINE",
                          "header_link": {
                              "mobile_web_url": "https://developers.kakao.com",
                              "web_url": "https://developers.kakao.com"
                          },
                          "contents": [
                              {
                                  "title": "취미의 특징, 탁구",
                                  "description": "스포츠",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/bDPMIb/btqgeoTRQvd/49BuF1gNo6UXkdbKecx600/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "크림으로 이해하는 커피이야기",
                                  "description": "음식",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/QPeNt/btqgeSfSsCR/0QJIRuWTtkg4cYc57n8H80/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "감성이 가득한 분위기",
                                  "description": "사진",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/c7MBX4/btqgeRgWhBy/ZMLnndJFAqyUAnqu4sQHS0/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ],
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"shareDefault(templatable:) - location", action: {
                    
                    let locationTemplateJsonStringData =
                      """
                      {
                          "object_type": "location",
                          "address_title": "카카오 판교오피스 카페톡",
                          "address": "경기 성남시 분당구 판교역로 235 에이치스퀘어 N동 8층",
                          "content": {
                              "description": "이번 주는 체리블라썸라떼 1+1",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/bSbH9w/btqgegaEDfW/vD9KKV0hEintg6bZT4v4WK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.com"
                              },
                              "title": "신메뉴 출시❤️ 체리블라썸라떼"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          }
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"shareDefault(templatable:) - commerce", action: {
                    
                    let commerceTemplateJsonStringData =
                      """
                      {
                          "object_type": "commerce",
                          "commerce": {
                              "product_name": "Ivory long dress",
                              "regular_price": 208800,
                              "discount_price": 146160,
                              "discount_rate": 30,
                              "currency_unit": "₩",
                              "currency_unit_position": 1
                          },
                          "content": {
                              "title": "Ivory long dress (4 Color)",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/RY8ZN/btqgOGzITp3/uCM1x2xu7GNfr7NS9QvEs0/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              },
                          },
                          "buttons": [
                              {
                                  "title": "구매하기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "공유하기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"shareDefault(templatable:) - text", action: {
                    let text = "카카오링크는 카카오 플랫폼 서비스의 대표 기능으로써 사용자의 모바일 기기에 설치된 카카오 플랫폼과 연동하여 다양한 기능을 실행할 수 있습니다.\\n\\n현재 이용할 수 있는 카카오링크는 다음과 같습니다.\\n카카오톡링크\\n카카오톡을 실행하여 사용자가 선택한 채팅방으로 메시지를 전송합니다.\\n카카오스토리링크\\n카카오스토리 글쓰기 화면으로 연결합니다."
                    let textTemplateJsonStringData =
                    """
                    {
                        "object_type": "text",
                        "text": "\(text)",
                        "link": {
                            "web_url": "https://developers.kakao.com",
                            "mobile_web_url": "https://developers.kakao.com"
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    else {
                        print("templatable is nil")
                    }
                }),
                Row(title:"shareDefault(templatable:) - calendar", action: {
                    let calendarTemplateJsonStringData =
                      """
                        {
                            "object_type": "calendar",
                            "id_type": "\(CalendarTemplate.IdType.Calendar.rawValue)",
                            "id": "\(AppDelegate.calendarId().calendar)",
                            "content": {
                                "link": {
                                    "web_url": "https://devtalk.kakao.com",
                                    "mobile_web_url": "https://developers.kakao.com"
                                },
                                "title": "2022 카타르 월드컵",
                                "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                            },
                            "buttons": [
                                {
                                    "title": "월드컵 일정 보기",
                                    "link": {
                                        "mobile_web_url": "https://developers.kakao.com",
                                        "web_url": "https://developers.kakao.com"
                                    }
                                }
                            ]
                        }
                      """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    else {
                        print("templatable is nil")
                    }
                }),
                Row(title:"shareDefault(templatable:) - calendar(event)", action: {
                    let calendarTemplateJsonStringData =
                      """
                      {
                          "object_type": "calendar",
                          "id_type": "\(CalendarTemplate.IdType.Event.rawValue)",
                          "id": "\(AppDelegate.calendarId().event)",
                          "content": {
                              "link": {
                                  "web_url": "https://devtalk.kakao.com",
                                  "mobile_web_url": "https://developers.kakao.com"
                              },
                              "title": "1월 신작 평론 모임",
                              "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                          },
                          "buttons": [
                              {
                                  "title": "모임 주제 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                        ShareApi.shared.rx.shareDefault(templatable:templatable)
                            .subscribe(onSuccess: { (sharingResult) in
                                self.success()
                                
                                //do something
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    else {
                        print("templatable is nil")
                    }
                }),
                Row(title:"imageUpload(image:)", action: {
                    if let image = UIImage(named: "sample1") {
                        ShareApi.shared.rx.imageUpload(image:image)
                            .subscribe(onSuccess: { (imageUploadResult) in
                                self.success()
                                
                                //do something
                                _ = imageUploadResult
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"imageScrap(image:)", action: {
                    if let url = URL(string:"https://t1.kakaocdn.net/kakaocorp/Service/KakaoTalk/pc/slide/talkpc_theme_01.jpg") {
                        ShareApi.shared.rx.imageScrap(imageUrl: url)
                            .subscribe(onSuccess: { (imageUploadResult) in
                                self.success()
                                
                                //do something
                                _ = imageUploadResult
                                
                            }, onFailure: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                })
            ]),
            // MARK: - section kakao navi
            SectionOfRow(header: "KAKAONAVI", items: [
                Row(title: "share(destination:) - katec", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "321286", y: "533707")
                    guard let shareUrl = NaviApi.shared.shareUrl(destination: destination) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(shareUrl) {
                        UIApplication.shared.open(shareUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                }),
                Row(title: "share(destination:) - wgs84", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "127.108640", y: "37.402111")
                    let option = NaviOption(coordType: .WGS84)
                    guard let shareUrl = NaviApi.shared.shareUrl(destination: destination, option: option) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(shareUrl) {
                        UIApplication.shared.open(shareUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                }),
                Row(title: "navigate(destination:viaList:) - katec", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "321286", y: "533707")
                    let viaList = [NaviLocation(name: "판교역 1번출구", x: "321525", y: "532951")]
                    guard let navigateUrl = NaviApi.shared.navigateUrl(destination: destination, viaList: viaList) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(navigateUrl) {
                        UIApplication.shared.open(navigateUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                }),
                Row(title: "navigate(destination:viaList:) - wgs84", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "127.108640", y: "37.402111")
                    let option = NaviOption(coordType: .WGS84)
                    let viaList = [NaviLocation(name: "판교역 1번출구", x: "127.111492", y: "37.395225")]
                    guard let navigateUrl = NaviApi.shared.navigateUrl(destination: destination, option:option, viaList: viaList) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(navigateUrl) {
                        UIApplication.shared.open(navigateUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                })
            ]),
            // MARK: - section kakaosync
            SectionOfRow(header: "KAKAOSYNC", items: [
                 Row(title: "약관선택 로그인", action: {
                    
                    let serviceTerms = ["service", "policy"]
                    
                    UserApi.shared.rx.loginWithKakaoAccount(serviceTerms: serviceTerms)
                    .subscribe(onNext: { (oauthToken) in
                        self.success()
                        
                        //do something
                        _ = oauthToken
                        
                    }, onError: {error in
                        self.errorHandler(error: error)
                    })
                    .disposed(by: self.disposeBag)
                 })
            ]),
            // MARK: - section etc
            SectionOfRow(header: "ETC", items: [
                Row(title: "Set custom token_manager ", action: {
                    print(">> Set custom token manager ")
                    Auth.shared.setTokenManager(TestCustomTokenManager())
                }),
                Row(title: "Check custom token_manager token ", action: {
                    var found = false
                    for (key, value) in UserDefaults.standard.dictionaryRepresentation() {
                        if (key == TestCustomTokenManager.tokenKey) {
                            print(">> custom token_manager token: \(String(describing:value))\n")
                            found = true
                            break
                        }
                    }
                    
                    if (found == false) {
                        print(">> not exist custom token_manager token")
                    }
                }),
                Row(title: "Set default token_manager ", action: {
                    print(">> Set default token manager ")
                    Auth.shared.setTokenManager()
                }),
                Row(title: "hasToken() usage", action: {
                    let disposeBag = DisposeBag()
                    
                    if (AuthApi.hasToken()) {
                        UserApi.shared.rx.accessTokenInfo()
                            .subscribe(onSuccess:{ (_) in
                                //토큰 유효성 체크 성공(필요시 토큰갱신됨)
                            }, onFailure: {error in
                                if let sdkError = error as? SdkError, sdkError.isInvalidTokenError() == true  {
                                   //login 필요
                                }
                                else {
                                    //기타에러.
                                }
                            })
                            .disposed(by: disposeBag)
                    }
                    else {
                        //login 필요
                    }
                })
            ])
            // section etc start =============================================================================================
        ]
        
    }

}


class TestCustomTokenManager : TokenManagable {
    static let tokenKey = "test_token_key"
    
    
    func getToken() -> OAuthToken? {
        print(">> TestCustomTokenManager.getToken() called.")
        
        if let data = UserDefaults.standard.data(forKey: TestCustomTokenManager.tokenKey) {
            let token = try? JSONDecoder().decode(OAuthToken.self, from:data)
            print(">> loaded token: \(String(describing:token))\n\n")
            
            return token
        }
        return nil
    }
    
    func setToken(_ token: OAuthToken?) {
        print(">> TestCustomTokenManager.setToke(_) called. token: \(String(describing:token)) ")
        
        if let data = try? JSONEncoder().encode(token) {
            print(">> 토큰 암호화를 권장합니다.\n\n")
            
            UserDefaults.standard.set(data, forKey:TestCustomTokenManager.tokenKey)
            UserDefaults.standard.synchronize()
        }
    }
    
    func deleteToken() {
        print(">> TestCustomTokenManager.deleteToken() called.")
        
        UserDefaults.standard.removeObject(forKey: TestCustomTokenManager.tokenKey)
    }
}
